# Copyright (c) 2002-2003 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2002 CIRB, Belgique
# Author: Julien Jalon <mailto:jj@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: LDAPDirectory.py,v 1.15 2003/09/25 13:55:19 anguenot Exp $

from string import maketrans

import ldap

from zLOG import LOG, INFO, DEBUG
from Globals import InitializeClass, DTMLFile

from AccessControl import ClassSecurityInfo
from BaseDirectory import BaseDirectory, DirectoryEditPermission, \
    MalformedProperties

try:
    # from LDAPUserFolder
    from Products.LDAPUserGroupsFolder.utils import _createLDAPPassword
except ImportError:
    LOG('NuxMetaDirectories', INFO, "Can't use LDAPUserGroupsFolder's _createLDAPPassword - using simple one")
    def _createLDAPPassword(prop):
        return prop

class LDAPDirectory(BaseDirectory):
    """
    LDAP directory
    """

    meta_type = 'LDAP Directory'

    host = ''
    port = 389
    user = ''
    password = ''
    baseDN = ''
    objectClass = '*'

    security = ClassSecurityInfo()

    manage_options = (BaseDirectory.manage_options[:2] +
                        (
                            {'label': 'LDAP',
                             'action': 'manage_editldapForm'
                             },
                         ) +
                    BaseDirectory.manage_options[2:])

    # BaseDirectory
    def _getEntry(self, entry_id, mtool=None):
        #LOG('NMD', INFO, '_getEntry LDAP: entry id is %s' % entry_id)
        entry_id = self.latin1_to_utf8(entry_id)
        ldap_prop = self._getInternalEntryProp()
        ldap_propids = self._getSchemaInternalKeys()
        if ldap_prop == 'dn':
            dn = self._normalizeDN(entry_id)
            if not dn.endswith(self.baseDN):
                return None
            baseDN = entry_id
            scope = ldap.SCOPE_BASE
            query_string = 'objectClass=%s' % self.objectClass
        else:
            baseDN = self.baseDN
            scope = ldap.SCOPE_SUBTREE
            query_string = '(&(objectClass=%s)(%s=%s))' % (self.objectClass, ldap_prop, entry_id)
        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        try:
            entries = cnx.search_s(baseDN, scope, query_string, ldap_propids)
        except:
            LOG('NMD', INFO, 'Error while getting entry')
            entries = []
        cnx.unbind_s()
        if not entries:
            #LOG('NMD', INFO, '_getEntry LDAP: no entry')
            return None
        entry = entries[0]
        #LOG('NMD', INFO, '_getEntry LDAP: LDAP entry is %s' % (entry, ))
        return self._decodeEntryForReading(entry)

    def _searchEntry(self, search_options, **kw):
        #LOG('NMD', INFO, '_searchEntry LDAP: %s' % kw)
        if search_options is not None:
            ldapscope = search_options.get('scope', 'sub')
            if ldapscope == 'base':
                scope = ldap.SCOPE_BASE
            elif ldapscope == 'one':
                scope = ldap.SCOPE_ONELEVEL
            else:
                scope = ldap.SCOPE_SUBTREE
            search_op = search_options.get('op', 'and')
            if search_op == 'or':
                op = '|'
            else:
                op = '&'
            search_type = search_options.get('type', 'in')
        else:
            scope = ldap.SCOPE_SUBTREE
            op = '&'
            search_type = 'in'

        searchkeys = kw.keys()
        query_list = ['(objectClass=%s)' % (self.objectClass)]
        baseDN = self.baseDN

        if 'baseDN' in searchkeys:
            new_base = self._normalizeDN(kw['baseDN'])
            if not new_base.endswith(baseDN):
                return []
            baseDN = new_base
            del kw['baseDN']
        baseDN = self.latin1_to_utf8(baseDN)

        if 'dn' in searchkeys:
            dn = self.latin1_to_utf8(self._normalizeDN(kw['dn']))
            del kw['dn']
        else:
            dn = None
            
        searchkeys = [key for key in searchkeys if key not in ('dn', 'baseDN')]
        list_type = type([])
        tuple_type = type(())
        string_type = type('')
        for key in searchkeys:
            criterium = kw[key]
            crit_type = type(criterium)
            if crit_type is string_type:
                if search_type == 'exact':
                    search_crit = key + '=' + self.latin1_to_utf8(criterium)
                else:
                    search_crit = key + '=*' + self.latin1_to_utf8(criterium) + '*'
                query_list.append('(' + search_crit + ')')
            elif crit_type is list_type or crit_type is tuple_type:
                for crit_value in criterium:
                    query_list.append('(' + key + '=' + self.latin1_to_utf8(crit_value) + ')')
            else:
                query_list.append('(%s=%s)' % (key, criterium))
        query_string = '(' + op + (''.join(query_list)) + ')'
        #LOG('NMD', INFO, '_searchEntry LDAP: query string: %s' % query_string)
        #LOG('NMD', INFO, '_searchEntry LDAP: baseDN: %s' % baseDN)
        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        try:
            entries = cnx.search_s(baseDN, scope, query_string, self._getSchemaInternalKeys())
        except:
            LOG('NMD', INFO, 'Error while searching entries')
            entries = []
        metadirectory_entries = [self._decodeEntryForReading(entry) for entry in entries]
        if dn is not None:
            found = []
            for entry in metadirectory_entries:
                if self._normalizeDN(entry['dn']) == dn:
                    found.append(entry)
            metadirectory_entries = found
        return metadirectory_entries

    def _createEntry(self, entry_id, entry):
        alias_prop = self._getInternalEntryProp()
        if alias_prop == 'dn':
            dn = self._checkDN(entry_id, self.baseDN)
            del entry['dn']
        else:
            if entry.has_key('baseBN'):
                baseDN = self._checkDN(entry['baseDN'], self.baseDN)
            else:
                baseDN = self.baseDN
            dn = '%s=%s,%s' % (alias_prop, entry[alias_prop], baseDN)
        if entry.has_key('baseDN'):
            del entry['baseDN']
        dn = self.latin1_to_utf8(dn)
        modlist = []
        entry = self._encodeEntryForWriting(entry)
        for key in entry.keys():
            prop = entry[key]
            if prop:
                modlist.append((key, prop))
        if self.objectClass == '*':
            modlist.append(('objectClass', ['top']))
        else:
            modlist.append(('objectClass', ['top', self.objectClass]))
        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        cnx.add_s(dn, modlist)
        cnx.unbind_s()

    def _writeEntry(self, entry_id, entry):
        #LOG('NMD', INFO, 'Will write in ldap: %s' % entry)
        alias_prop = self._getInternalEntryProp()
        if alias_prop == 'dn':
            dn = self._checkDN(entry_id, self.baseDN)
        else:
            dn = '%s=%s,%s' % (alias_prop, entry_id, self.baseDN)
        if entry.has_key('dn'):
            del entry['dn']
        if entry.has_key('baseDN'):
            del entry['baseDN']
        if entry.has_key(alias_prop):
            del entry[alias_prop]

        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        try:
            ldap_entry = self._decodeEntryForReading(cnx.search_s(dn, ldap.SCOPE_BASE, 'objectClass=' + self.objectClass)[0], include_dns=0)
        except:
            LOG('NMD', INFO, 'Error while getting old LDAP entry')
            cnx.unbind_s()
            raise
        cnx.unbind_s()
        modlist = []
        entry = self._encodeEntryForWriting(entry)
        for key in entry.keys():
            prop = entry[key]
            if prop:
                if ldap_entry.has_key(key):
                    modlist.append((ldap.MOD_REPLACE, key, prop))
                else:
                    modlist.append((ldap.MOD_ADD, key, prop))
            else:
                if ldap_entry.has_key(key):
                    modlist.append((ldap.MOD_DELETE, key, None))
        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        try:
            cnx.modify_s(dn, modlist)
        except:
            cnx.unbind_s()
            LOG('NMD', INFO, 'Error while writing entry')
            raise
        cnx.unbind_s()

    def _deleteEntry(self, entry_id):
        alias_prop = self._getInternalEntryProp()
        if alias_prop == 'dn':
            dn = self._checkDN(entry_id, self.baseDN)
        else:
            dn = '%s=%s,%s' % (alias_prop, entry_id, self.baseDN)
        dn = self.latin1_to_utf8(dn)
        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        try:
            self._delete(dn, cnx)
        except:
            LOG('NMD', INFO, 'Error while deleting entry')
            cnx.unbind_s()
            raise
        cnx.unbind_s()

    def _getNewEntry(self, entry_id):
        if 'baseDN' in self._getSchemaInternalKeys():
            return {'baseDN': self.baseDN,
                    self._getInternalEntryProp(): entry_id}
        else:
            return {self._getInternalEntryProp(): entry_id}

    # public utilities

    security.declarePublic('createDNFromEntry')
    def createDNFromEntry(self, entry, prop):
        baseDN = entry.get('baseDN', None)
        if baseDN is None:
            baseDN = self._normalizeDN(self.baseDN)
        else:
            if not self._normalizeDN(baseDN).endswith(self._normalizeDN(self.baseDN)):
                raise MalformedProperties, {'baseDN': baseDN}
        prop_value = entry.get(prop, None)
        if not prop_value:
            raise MalformedProperties, {prop: 'Required'}
        new_dn = '%s=%s,%s' % (prop, prop_value, baseDN)
        return new_dn

    # utilities
    security.declarePrivate('sanitize_entry')
    def sanitize_entry(self, entry):
        """
        Utility to avoid the problem of ids that might contain
        latin-1 encoding chars.
        Like with M$ Active Directory.
        """
        entry = entry or ''
    	entry = entry[max(entry.rfind('/'),
                                entry.rfind('\\'),
                                entry.rfind(':'))+1:]
        entry = entry.replace('%/1iso8859-15æ', 'ae')
        # Losing accents may not be necessary
    	tr = maketrans('',
    	               'AAAAAACEEEEIIIINOOOOOOUUUUYaaaaaaceeeeiiiinoooooouuuuyy')
    	entry = entry.translate(tr)
    	ok = ('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
              ' -_.,!?()[]&%$' "'") # at least not = ; / \ :
    	entry = ''.join([c for c in entry if c in ok])
    	entry = entry.strip()
    	return entry
    
    security.declarePrivate('utf8_to_latin1')
    def utf8_to_latin1(self, s):        
        try:
            return unicode(s, 'UTF-8').encode('latin1')
        except UnicodeError:
            LOG("CONVERTION UTF-8 FAILED : LATIN-1 entry : Should be a M$ stuff", DEBUG, s)
            return unicode(self.sanitize_entry(s), 'UTF-8').encode('latin1')

    security.declarePrivate('latin1_to_utf8')
    def latin1_to_utf8(self, s):
        return unicode(s, 'latin1').encode('UTF-8')

    def _encodeEntryForWriting(self, metadirectory_entry):
        #LOG('NMD', INFO, 'Encoding for writing: %s' % metadirectory_entry)
        ldap_entry = {}
        for key in self._schemakeys:
            property = self._schema[key]
            if property.isexternal:
                continue
            alias = property.getInternalName()
            type = property.type
            prop = metadirectory_entry.get(alias)
            if prop is None:
                continue
            #LOG('NMD', INFO, 'Encoding property %s' % alias)
            if type == 'string':
                ldap_prop = self.latin1_to_utf8(str(prop))
            elif type == 'list':
                ldap_prop = [self.latin1_to_utf8(str(item)) for item in prop]
            elif type == 'password':
                # XXX here, ldap directory must now what type of password
                # encoding he has to know
                ldap_prop = _createLDAPPassword(prop)
            elif type == 'binary':
                ldap_prop = prop
            ldap_entry[alias] = ldap_prop
        #LOG('NMD', INFO, 'Encoded entry: %s' % ldap_entry)
        return ldap_entry

    def _encodeProperty(self, prop, ts=type(''), tl=type([]), tn=type(())):
        prop_type = type(prop)
        if prop_type is ts:
            return self.latin1_to_utf8(prop)
        elif prop_type is ts or prop_type is tn:
            return [self.latin1_to_utf8(item) for item in prop]
        else:
            return prop

    def _decodeEntryForReading(self, entry, include_dns=1):
        ldap_entry = entry[1]
        entrykeys = ldap_entry.keys()
        metadirectory_entry = {}
        for key in self.getSchemaKeys():
            property = self._schema[key]
            if property.isexternal:
                continue
            alias = property.getInternalName()
            if include_dns:
                if alias == 'dn':
                    metadirectory_entry['dn'] = entry[0]
                    continue
                if alias == 'childDNs':
                    metadirectory_entry['childDNs'] = self._childDNs(entry[0])
                if alias == 'baseDN' and entry[0] != self.baseDN:
                    metadirectory_entry['baseDN'] = ','.join([rdn.strip() for rdn in entry[0].split(',')[1:]])
            if alias in entrykeys:
                prop = ldap_entry[alias]
                if property.type == 'string':
                    metadirectory_entry[alias] = self.utf8_to_latin1(','.join(prop))
                elif property.type == 'list':
                    metadirectory_entry[alias] = [self.utf8_to_latin1(item) for item in prop]
                else:
                    metadirectory_entry[alias] = ','.join(prop)
        return metadirectory_entry

    def _checkDN(self, dn, baseDN):
        if self._normalizeDN(dn).endswith(self._normalizeDN(baseDN)):
            return dn
        raise ValueError, 'dn=%s and baseDN=%s' % (dn, baseDN)

    def _normalizeDN(self, dn):
        return ','.join([rdn.strip() for rdn in dn.split(',')])

    def _delete(self, dn, cnx):
        #LOG('NMD', INFO, 'deleting dn %s' % dn)
        childs = cnx.search_s(dn, ldap.SCOPE_ONELEVEL, 'objectClass=*', ['dn'])
        dns = [child[0] for child in childs]
        for child_dn in dns:
            self._delete(child_dn, cnx)
        cnx.delete(dn)

    def _childDNs(self, dn):
        query_string = 'objectClass=' + self.objectClass
        entry_alias = self._getInternalEntryProp()
        cnx = ldap.open(self.host, int(self.port))
        cnx.simple_bind_s(self.user, self.password)
        try:
            childs = cnx.search_s(dn, ldap.SCOPE_ONELEVEL, query_string, [entry_alias])
        except:
            childs = []
        if entry_alias == 'dn':
            metadirectory_childs = [child[0] for child in childs]
        else:
            metadirectory_childs = [child[1][entry_alias] for child in childs]
        return metadirectory_childs

    # ZMI

    manage_editldapForm = DTMLFile('zmi/manage_editldapForm', globals())

    security.declareProtected(DirectoryEditPermission, 'manage_editldap')
    def manage_editldap(self, host=None, port=None, user=None, password=None,
                        baseDN=None, objectClass=None, REQUEST=None):
        """
            Edit LDAP properties
        """
        if host is not None:
            self.host = host
        if port is not None:
            self.port = port
        if user is not None:
            self.user = user
        if password is not None:
            self.password = password
        if baseDN is not None:
            self.baseDN = baseDN
        if objectClass is not None:
            self.objectClass = objectClass
        if REQUEST is not None:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '/manage_editldapForm?manage_tabs_message=Changed.')

InitializeClass(LDAPDirectory)

manage_addLDAPDirectoryForm = DTMLFile('zmi/manage_addLDAPDirectoryForm', globals())

def manage_addLDAPDirectory(dispatcher, id, title='', REQUEST=None):
    """
        Add a LDAP Directory
    """
    ob = LDAPDirectory(id, title=title)
    container = dispatcher.Destination()
    container._setObject(id, ob)
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
