# Copyright (c) 2002-2003 Nuxeo SARL <http://nuxeo.com>
# Copyright (c) 2002 CIRB, Belgique
# Author: Julien Jalon <mailto:jj@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: MembersDirectory.py,v 1.26 2003/12/13 12:15:54 fermigier Exp $

import random

from zLOG import LOG, DEBUG, ERROR
from Globals import InitializeClass, DTMLFile, Persistent
import Acquisition
from Acquisition import aq_inner, aq_parent

from AccessControl import ClassSecurityInfo, Unauthorized
from Products.CMFCore.utils import getToolByName
from BaseDirectory import BaseDirectory, DirectoryEditPermission, \
    DirectorySchemaProperty

USERIDPROPID = "id_"
PASSWORDPROPID = "password_"
ROLESPROPID = "roles_"
GROUPSPROPID = "groups_"

class MembersNomenclatures(Persistent, Acquisition.Implicit):

    security = ClassSecurityInfo()

    security.declarePublic("roles")
    def roles(self):
        """return roles"""
        return {'Manager': '_role_Manager_', 'Member': '_role_Member_'}

    security.declarePublic("groups")
    def groups(self):
        """return groups"""
        aclu = aq_inner(self).acl_users
        if hasattr(aclu, 'getGroupNames'):
            groupids = aclu.getGroupNames()
        else:
            groupids = []
        groups = {}
        for groupid in groupids:
            groups[groupid] = aclu.getGroupById(groupid).Title() or groupid
        return groups

InitializeClass(MembersNomenclatures)

class MembersDirectory(BaseDirectory):
    """
    Members directory
    uses UserFolder, MemberData and Membership tools
    """

    meta_type = 'Members Directory'

    security = ClassSecurityInfo()

    manage_options = (BaseDirectory.manage_options[:2] +
                        (
                            {'label': 'Synchronize',
                             'action': 'manage_synchronize'
                             },
                         ) +
                    BaseDirectory.manage_options[2:])

    _nosync = (
        'portal_skin',
        'listed',
        'login_time',
        'last_login_time',
        )


    def __init__(self, id, **kw):
        BaseDirectory.__init__(self, id, **kw)
        self.manage_properties(
            acl_write="python:manager or member.getUserName()==entry_id",
            acl_create="manager", acl_delete="python:manager and entry_id != member.getUserName()",
            entry_prop=USERIDPROPID, display_prop='fullname')
        self.nomenclatures = MembersNomenclatures()
        self._schemakeys = (USERIDPROPID, 'fullname', PASSWORDPROPID, ROLESPROPID, GROUPSPROPID, 'name', 'givenName', 'email')
        self._schema = {
            USERIDPROPID: DirectorySchemaProperty(USERIDPROPID, title="_dir_prop_Login_",
                                alias=USERIDPROPID, acl_write="python:not property or nobody", process_write="python:property or malformed('Required')"),
            PASSWORDPROPID: DirectorySchemaProperty(PASSWORDPROPID, title="_dir_prop_Password_",
                        alias=PASSWORDPROPID, acl_read='nobody',
                        acl_write="python:manager or member.getUserName()==entry_id",
                        process_write="property|nothing",
                        type="password", issearchable=0),
            ROLESPROPID: DirectorySchemaProperty(ROLESPROPID, title="_dir_prop_Roles_",
                                alias=ROLESPROPID, acl_write="manager",
                                process_write="python:('Member' not in property) and (property + ['Member']) or property",
                                nomenclature="nomenclatures/roles",
                                type="list"
                           ),
            GROUPSPROPID: DirectorySchemaProperty(GROUPSPROPID, title="_dir_prop_Groups_",
                                alias=GROUPSPROPID, acl_write="manager",
                                nomenclature="nomenclatures/groups",
                                type="list", isdirectory_ref=1, directory_ref='groups'),
            'name': DirectorySchemaProperty('name', title="_dir_prop_Name_", alias="sn",
                acl_write="python:manager or member.getUserName() == entry_id"),
            'givenName': DirectorySchemaProperty('givenName', title="_dir_prop_givenName_",
                alias="givenName",
                acl_write="python:manager or member.getUserName() == entry_id"),
            'email': DirectorySchemaProperty('email', title="_dir_prop_E-mail_",
                alias="email",
                acl_write="python:manager or member.getUserName() == entry_id"),
            'fullname': DirectorySchemaProperty('fullname', title="_dir_prop_fullName_",
                acl_write="nobody", process_read="python:(entry.get('name',None) or entry.get('givenName',None)) and (entry.get('givenName','') + ' ' + entry.get('name','')) or entry_id", isexternal = 1,
                issearchable = 0, isvisible=0)
                    }


    # BaseDirectory
    def _getEntry(self, entry_id, mtool=None):
        if mtool is None:
            mtool = getToolByName(self, 'portal_membership')

        # XXX: is this test really necessary ?
        mdtool = getToolByName(self, 'portal_memberdata')
        if entry_id not in mdtool._members.keys():
            return None

        member = mtool.getMemberById(entry_id)
        if member is None:
            return None
        roles = member.getRoles()
        user = self._getUser(member)
        aclu = aq_parent(aq_inner(user))

        if hasattr(aclu, 'setGroupsOfUser'):
            try:
                groups = user.getGroups()
            except AttributeError:
                groups = []
        else:
            groups = None

        rolenom = self._schema['roles_'].__of__(self).getNomenclature()
        if rolenom is None:
            okroles = self.nomenclatures.roles().keys()
        else:
            okroles = rolenom.keys()
        roles = [role for role in roles if role in okroles]
        entry = {}
        entry[USERIDPROPID] = entry_id
        entry[PASSWORDPROPID] = ''
        entry[ROLESPROPID] = roles
        if groups is not None:
            entry[GROUPSPROPID] = groups
        for propid in self._schemakeys:
            property = self._schema[propid]
            alias = property.getInternalName()
            if entry.has_key(propid):
                continue
            try:
                value = member.getProperty(alias, None)
            except:
                value = None
            if value is None:
                continue
            if property.type == 'list':
                if type(value) is type(''):
                    value = value.split(',')
                    value = [item.strip() for item in value]
            entry[alias] = value
        return entry

    def _searchEntry(self, search_options, **kw):
        mtool = getToolByName(self, 'portal_membership')

        if USERIDPROPID in kw.keys():
            value = kw[USERIDPROPID]
            del kw[USERIDPROPID]
            kw['id'] = value

        ids = self.searchMemberData(kw, search_options)

        return filter(None, [self._getEntry(id, mtool=mtool) for id in ids])

    def _createEntry(self, entry_id, entry):
        if not entry_id:
            return
        mtool = getToolByName(self, 'portal_membership')
        password = self._randomPassword()
        roles = ()
        domains = []
        # XXX this should check that the member doesn't already exist
        mtool.addMember(entry_id, password, roles, domains)
        member = mtool.getMemberById(entry_id)
        if member is None:
            raise ValueError, "Couldn't add member"
        self._writeEntry(entry_id, entry)

    def _writeEntry(self, entry_id, entry):
        LOG('MembersDirectory._writeEntry', DEBUG, 'entry_id=%s' % entry_id)
        mtool = getToolByName(self, 'portal_membership')
        member = mtool.getMemberById(entry_id)
        if member is None:
            return
        mapping = {}
        for propid, value in entry.items():
            if propid == USERIDPROPID:
                continue
            if propid == PASSWORDPROPID:
                password = value
                if password != '':
                    self._setMemberPassword(member, password)
                continue
            if propid == ROLESPROPID:
                roles = value
                self._setMemberRoles(member, roles)
                continue
            if propid == GROUPSPROPID:
                groups = value
                done = self._setMemberGroups(member, groups)
                if done: continue
                # else we treat it as a normal property
            mapping[propid] = value

        member.setMemberProperties(mapping)

    def _deleteEntry(self, entry_id):
        mtool = getToolByName(self, 'portal_membership')
        aclu = mtool.acl_users
        member = mtool.getMemberById(entry_id)
        if member is not None:
            aclu.userFolderDelUsers([entry_id])
        mdtool = getToolByName(self, 'portal_memberdata')
        mdtool.pruneMemberDataContents()

    def _getNewEntry(self, entry_id):
        new_entry = { USERIDPROPID: entry_id,
                      ROLESPROPID: ('Member', ),
                      GROUPSPROPID: (),
                    }
        return new_entry

    # utilities

    def _setMemberPassword(self, member, password):
        """Sets a member's password."""
        user = self._getUser(member)
        aclu = aq_parent(aq_inner(user))
        name = user.getUserName()
        aclu._doChangeUser(name, password, member.getRoles(), 
                           member.getDomains())

    def _setMemberRoles(self, member, roles):
        """Sets a member's roles."""
        user = self._getUser(member)
        aclu = aq_parent(aq_inner(user))
        name = user.getUserName()
        aclu._doChangeUser(name, None, roles, member.getDomains())

    def _setMemberGroups(self, member, groups):
        """Set a member's groups.

        Returns true if it could be set in a user folder
        """
        user = self._getUser(member)
        aclu = aq_parent(aq_inner(user))
        if hasattr(aclu, 'setGroupsOfUser'):
            userid = user.getUserName()
            aclu.setGroupsOfUser(groups, userid)
            return 1
        return 0

    def _getUser(self, member):
        """Gets the user from a member. Treats anonymous user."""
        if member.getUserName() == 'Anonymous User':
            user = member
        else:
            try:
                # XXX In some unknown cases this apparently gives:
                # TypeError: getUser() takes exactly 2 arguments (1 given)
                user = member.getUser()
            except TypeError:
                try:
                    cls = member.__class__
                except:
                    cls = '(no class)'
                LOG('MembersDirectory._getUser', ERROR,
                    'Got TypeError, member=%s class=%s' % (member, cls))
                raise
        return user

    def _randomPassword(self, len=20):
        """Computes a random password."""
        chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
        pwd = ''
        for i in range(len):
            pwd += random.choice(chars)
        return pwd

    def searchMemberData(self, params, search_options={}):
        """Searches member using MemberData._members.

        Does an AND search for all key, value of the params
        If the member attributes corresponding to a key is a list,
        does an OR search on all the list elements.
        If the passed value is a string, does a substring lowercase.
        If the passed value is a list, does OR search with exact match.
        Returns a list of member ids.

        This function accepts an optional search_options dictionary parameter.
        As of now, recognized search_options keys are :
          * exact_match with values 0 and 1
        """

        _marker = []
        res = []
        inlist = {}
        searchlist = {}

        for key, value in params.items():
            if type(value) is type(''):
                inlist[key] = 0
                searchlist[key] = value.lower()
            else:
                inlist[key] = 1
                searchlist[key] = value
        mdtool = getToolByName(self, 'portal_memberdata')
        aclu = aq_inner(self).acl_users
        for user_wrapper in mdtool._members.values():
            # user_wrapper is a MemberData without context
            userid = user_wrapper.id
            user = _marker
            found = 1
            for key, value in searchlist.items():
                if key == ROLESPROPID:
                    if user is _marker: user = aclu.getUserById(userid)
                    if user is not None:
                        searched = user.getRoles()
                    else: # unpruned user in higher user folder
                        searched = None
                elif key == GROUPSPROPID:
                    if hasattr(aclu, 'setGroupsOfUser'):
                        if user is _marker: user = aclu.getUserById(userid)
                        if user is not None:
                            try:
                                searched = user.getGroups()
                            except AttributeError:
                                searched = None
                        else:
                            searched = None
                    else:
                        searched = None
                else:
                    searched = getattr(user_wrapper, key, None)
                if searched is None:
                    found = 0
                    break
                if type(searched) is type(''):
                    searched = (searched, )
                matched = 0
                for item in searched:
                    #LOG('MembersDirectory.searchMemberData', DEBUG,
                    #    'inlist / item / value = %s / %s / %s' % (inlist, item, value))
                    if inlist[key]:
                        matched = item in value
                    else:
                        exact_match = search_options.get('exact_match')
                        if exact_match:
                            matched = (item.lower() == value)
                        else:
                            matched = item.lower().find(value) != -1
                    if matched:
                        break
                if not matched:
                    found = 0
                    break
            if found:
                res.append(userid)

        return res


    def getMemberDataSchema(self):
        mdtool = getToolByName(self, 'portal_memberdata')
        # Core MemberData using properties
        propmap = mdtool._propertyMap()
        s_list = []
        s_dict = {}
        for d in propmap:
            id = d['id']
            title = d.get('label', id)
            t = d['type']
            is_list = (t == 'tokens' or
                       t == 'lines' or
                       t == 'multiple selection')
            s_list.append(id)
            s_dict[id] = {
                'title': title,
                'is_list': is_list,
                'nomenclature': '',
                }
        return (s_list, s_dict)

    security.declareProtected(DirectoryEditPermission, 'syncSchemaAndMemberData')
    def syncSchemaAndMemberData(self):
        """Synchronizes the schema and MemberData informations."""
        self.updateSchemaFromMemberData()
        self.updateMemberDataFromSchema()

    security.declarePrivate('updateSchemaFromMemberData')
    def updateSchemaFromMemberData(self):
        """Updates the schema from MemberData informations."""
        (dataschema_list, dataschema_dict) = self.getMemberDataSchema()
        schemaInternalList = [self._schema[id].getInternalName() for id in self._schemakeys]
        new_schemakeys = self._schemakeys
        for id in dataschema_list:
            # already in schema
            if id in schemaInternalList:
                continue
            # no sync for this one
            if id in self._nosync:
                continue
            d = dataschema_dict[id]
            new_prop = DirectorySchemaProperty(id, alias=id, title=d['title'],
                acl_write="python:manager or member.getUserName() == entry_id",
                type=(d['is_list'] and 'list' or 'string'))
            new_schemakeys = new_schemakeys + (id, )
            self._schema[id] = new_prop
        self._schemakeys = new_schemakeys

    security.declarePrivate('updateMemberDataFromSchema')
    def updateMemberDataFromSchema(self):
        """Updates the MemberData informations from schema"""
        mdtool = getToolByName(self, 'portal_memberdata')
        md_ids = [d['id'] for d in mdtool._propertyMap()]
        for propid in self._schemakeys:
            if propid in (USERIDPROPID, PASSWORDPROPID,
                                ROLESPROPID, GROUPSPROPID):
                continue
            property = self._schema[propid]
            if property.isexternal:
                continue
            id = property.getInternalName()
            if id in md_ids:
                continue
            type = (property.type == 'list') and 'tokens' or 'string'
            mdtool.manage_addProperty(id, '', type)

    security.declareProtected(DirectoryEditPermission, 'syncMembers')
    def syncMembers(self):
        """
        Synchronises new members from acl_users, and purge members that
        disappeared from acl_users
        """
        mtool = getToolByName(self, 'portal_membership')
        mdtool = getToolByName(self, 'portal_memberdata')
        members = mdtool._members
        member_list = members.keys()
        user_list = mtool.listMemberIds()
        # prune orphans
        orphans = [id for id in member_list if id not in user_list]
        LOG('NMD', DEBUG, 'deleting members %s' % (orphans, ))
        for id in orphans:
            self._deleteEntry(id)
        # sync new members from acl_users
        # We must make a new member_list from the updated members object,
        # in case ALL users have been deleted. Otherwise a segmentation fault
        # occurs
        member_list = members.keys()
        for id in user_list:
            LOG('NMD', DEBUG, 'testing acl user %s' % id)
            if id not in member_list:
                LOG('NMD', DEBUG, 'creating %s' % (id, ))
                mp = self.createEntry({self.entry_prop: id})

    security.declareProtected(DirectoryEditPermission, 'getNoSync')
    def getNoSync(self):
        return list(self._nosync)

    security.declareProtected(DirectoryEditPermission, 'getMiscInfo')
    def getMiscInfo(self):
        """
        Gets misc info.
        """
        # members
        mtool = getToolByName(self, 'portal_membership')
        mdtool = getToolByName(self, 'portal_memberdata')
        members = mdtool._members
        member_list = members.keys()
        user_list = mtool.listMemberIds()
        member_count = len(members)
        # members not in acl_users
        orphans = [id for id in member_list if id not in user_list]
        # member not registered in memberdata
        unregs = [id for id in user_list if id not in member_list]
        # schema items
        (dataschema_list, dataschema_dict) = self.getMemberDataSchema()
        mdschema = [id for id in dataschema_list if id not in self._nosync]
        schema = [self._schema[id].getInternalName()
                                    for id in self.getSchemaKeys() if not self._schema[id].isexternal]
        mdschemanotsync = [id for id in mdschema if id not in schema]
        schemanotsync = [id for id in schema if id not in mdschema+[USERIDPROPID, PASSWORDPROPID, ROLESPROPID, GROUPSPROPID]]

        return {
            'member_count': member_count,
            'orphans': orphans,
            'unregs': unregs,
            'mdschemanotsync': mdschemanotsync,
            'schemanotsync': schemanotsync,
            }

    # ZMI

    manage_synchronize = DTMLFile('zmi/manage_synchronize', globals())

    security.declareProtected(DirectoryEditPermission, 'manage_doSyncSchema')
    def manage_doSyncSchema(self, REQUEST=None):
        """Does the sync betweern the schema and MemberData."""
        self.syncSchemaAndMemberData()
        if REQUEST is not None:
            return self.manage_synchronize(self, REQUEST, manage_tabs_message="Schema synchronized.")

    security.declareProtected(DirectoryEditPermission, 'manage_doSyncMembers')
    def manage_doSyncMembers(self, REQUEST=None):
        """
        Sync the members between acl_users and MemberData
        """
        self.syncMembers()
        if REQUEST is not None:
            return self.manage_synchronize(self, REQUEST,
                manage_tabs_message="Members synchronized.")

    security.declareProtected(DirectoryEditPermission, 'manage_editNoSync')
    def manage_editNoSync(self, nosync, REQUEST=None, **kw):
        """Changes non-synchronized items."""
        l = []
        for n in nosync:
            if n and n not in l:
                l.append(n)
        self._nosync = l
        if REQUEST is not None:
            return self.manage_synchronize(self, REQUEST, manage_tabs_message="Changed.")

InitializeClass(MembersDirectory)

manage_addMembersDirectoryForm = DTMLFile('zmi/manage_addMembersDirectoryForm', globals())

def manage_addMembersDirectory(dispatcher, id, title='', REQUEST=None):
    """
        Add a Members Directory
    """
    ob = MembersDirectory(id, title=title)
    container = dispatcher.Destination()
    container._setObject(id, ob)
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)

GROUPIDPROPID = "id_"
GROUPTITLEPROPID = "title_"
GROUPMEMBERSPROPID = "members_"
GROUPGROUPSPROPID = "groups_"
_marker = []

class GroupsNomenclatures(Persistent, Acquisition.Implicit):

    security = ClassSecurityInfo()

    security.declarePublic("members")
    def members(self):
        """Return members."""

        members = self._getMembersDirectory()
        memberentryid = members.entry_prop
        memberdisplayid = members.display_prop
        memberslist = members.searchEntry()
        member_nomenclature = {}
        for member in memberslist:
            member_id = member[memberentryid]
            member_name = member.get(memberdisplayid, member_id)
            member_nomenclature[member_id] = member_name
        return member_nomenclature

    security.declarePublic("groups")
    def groups(self):
        """Return members."""

        groups = self._getGroupsDirectory()
        groupentryid = groups.entry_prop
        groupdisplayid = groups.display_prop
        groupslist = groups.searchEntry()
        group_nomenclature = {}
        for group in groupslist:
            group_id = group[groupentryid]
            group_name = group.get(groupdisplayid, group_id)
            group_nomenclature[group_id] = group_name
        return group_nomenclature

    def _getMembersDirectory(self):
        return aq_parent(aq_parent(aq_inner(self))).members

    def _getGroupsDirectory(self):
        return aq_parent(aq_parent(aq_inner(self))).groups

InitializeClass(GroupsNomenclatures)

class GroupsDirectory(BaseDirectory):
    """
    Groups directory
    uses UserGroupFolder
    """

    meta_type = "Groups Directory"

    security = ClassSecurityInfo()

    def __init__(self, id, **kw):
        BaseDirectory.__init__(self, id, **kw)
        self.manage_properties(
            acl_write="python:manager",
            acl_create="manager", acl_delete="python:manager",
            entry_prop=GROUPIDPROPID, display_prop=GROUPTITLEPROPID)
        self.nomenclatures = GroupsNomenclatures()
        self._schemakeys = (GROUPIDPROPID, GROUPTITLEPROPID, GROUPMEMBERSPROPID, )#GROUPGROUPSPROPID)
        self._schema = {
            GROUPIDPROPID: DirectorySchemaProperty(GROUPIDPROPID, title="_dir_prop_Id_",
                                alias=GROUPIDPROPID, acl_write="python:not property or nobody", process_write="python:property or malformed('Required')"),
            GROUPTITLEPROPID: DirectorySchemaProperty(GROUPTITLEPROPID, title="_dir_prop_groupName_", alias=GROUPTITLEPROPID),
            GROUPMEMBERSPROPID: DirectorySchemaProperty(GROUPMEMBERSPROPID, title="_dir_prop_Users_", alias=GROUPMEMBERSPROPID, nomenclature="nomenclatures/members", type="list", isdirectory_ref=1, directory_ref='members'),
            #GROUPGROUPSPROPID: DirectorySchemaProperty(GROUPGROUPSPROPID, title="_dir_prop_Groups_", alias=GROUPGROUPSPROPID, nomenclature="nomenclatures/groups", type="list", isdirectory_ref=1, directory_ref='groups', issearchable=0),
        }

    def _getEntry(self, entry_id):
        aclu = self.acl_users
        if not hasattr(aclu, 'getGroupById'):
            return None
        group = aclu.getGroupById(entry_id, None)
        if group is None:
            return None

        return self._getGroupEntry(group)

    def _getGroupMembers(self, groupobject):
        # Groups with groups will have two methods,
        # One getUsers() get all uses in the groups and the groups groups
        # and one getMembers() to get only the members of the group and
        # not the subgroups.
        # Groups that do not support subgroups may not have getMembers.
        if hasattr(groupobject, 'getMembers'):
            return groupobject.getMembers()
        if hasattr(groupobject, 'getUsers'):
            return groupobject.getUsers()
        return ()


    def _searchEntry(self, search_options, **kw):
        aclu = self.acl_users
        groups = aclu.getGroupNames()
        has_membersprop = self._hasMembersProp()

        idsearch = kw.get(GROUPIDPROPID, None)
        if idsearch is not None:
            idsearch = idsearch.lower()
        titlesearch = kw.get(GROUPTITLEPROPID, None)
        if titlesearch is not None:
            titlesearch = titlesearch.lower()
        memberssearch = kw.get(GROUPMEMBERSPROPID, [])
        entries = []
        for groupid in groups:
            ok = 1
            if idsearch is not None:
                if groupid.lower().find(idsearch) == -1:
                    ok = 0
            if ok:
                group = aclu.getGroupById(groupid)
                if titlesearch is not None:
                    if group.Title().lower().find(titlesearch) == -1:
                        ok = 0
                if ok and memberssearch:
                    users = self._getGroupMembers(group)
                    common = [user for user in memberssearch if user in users]
                    if common != memberssearch:
                        ok = 0
            if ok:
                entries.append(self._getGroupEntry(group, has_membersprop))
        return entries

    def _createEntry(self, entry_id, entry):
        if not entry_id:
            LOG('Members Directory', DEBUG, 'Can not create entry with empty id.')
            return
        aclu = self.acl_users
        usernames = entry.get(GROUPMEMBERSPROPID, [])
        title = entry.get(GROUPTITLEPROPID, '')
        if hasattr(aclu, 'userFolderAddGroup'):
            aclu.userFolderAddGroup(entry_id, title)
            aclu.setUsersOfGroup(usernames, entry_id)

        groups = entry.get(GROUPGROUPSPROPID, [])
        if groups:
            groupob = aclu.getGroupById(entry_id)
            if hasattr(groupob, 'setGroups'):
                groupob.setGroups(groups)


    def _writeEntry(self, entry_id, entry):
        aclu = self.acl_users

        group = aclu.getGroupById(entry_id, None)
        if group is None:
            raise ValueError, entry_id

        usernames = entry.get(GROUPMEMBERSPROPID, None)
        if usernames is not None:
            filtered_users = []
            for user in usernames:
                if aclu.getUserById(user):
                    filtered_users.append(user)
            aclu.setUsersOfGroup(filtered_users, entry_id)

        title = entry.get(GROUPTITLEPROPID, None)
        if title is not None:
            group.setTitle(title)

        groupids = entry.get(GROUPGROUPSPROPID, None)
        if groupids is not None and hasattr(aclu,'getGroupById'):
            filtered_groups = []
            for groupid in groupids:
                if aclu.getGroupById(groupid):
                    filtered_groups.append(groupid)
            groupob = aclu.getGroupById(entry_id)
            groupob.setGroups(filtered_groups)

    def _deleteEntry(self, entry_id):
        aclu = self.acl_users
        aclu.userFolderDelGroups([entry_id])

    def _getNewEntry(self, entry_id):
        return {GROUPIDPROPID: '', GROUPTITLEPROPID: '', GROUPMEMBERSPROPID: [],
                GROUPGROUPSPROPID: [],}

    # utilities

    def _hasMembersProp(self):
        has_membersprop = 0
        for prop in self._schemakeys:
            property_name = self._schema[prop].getInternalName()
            if property_name == GROUPMEMBERSPROPID:
                has_membersprop = 1
                break
        return has_membersprop

    def _getGroupEntry(self, group, has_membersprop=None):
        # has_membersprop is now ignored. Instead I test to figure
        # out if it has a members_prop and which it is.
        LOG('GroupDirectory', DEBUG, '_getGroupEntry')
        group_entry = {GROUPIDPROPID: group.id,
                       GROUPTITLEPROPID: group.Title()}
        getgroups = None
        group_entry[GROUPMEMBERSPROPID] = tuple(self._getGroupMembers(group))
        if hasattr(group, 'getGroups'):
            group_entry[GROUPGROUPSPROPID] = tuple(group.getGroups())
        return group_entry

InitializeClass(GroupsDirectory)

manage_addGroupsDirectoryForm = DTMLFile('zmi/manage_addGroupsDirectoryForm', globals())

def manage_addGroupsDirectory(dispatcher, id, title='', REQUEST=None):
    """
        Add a Groups Directory
    """
    ob = GroupsDirectory(id, title=title)
    container = dispatcher.Destination()
    container._setObject(id, ob)
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
