# Copyright (c) 2002-2003 Nuxeo SARL <http://nuxeo.com>
# Author: Julien Anguenot <ja@nuxeo.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#
# $Id: ZODBDirectory.py,v 1.4 2003/12/17 08:24:45 edegoute Exp $

__author__ ="Julien Anguenot <mailto:ja@nuxeo.com>"

"""
ZODBDirectory directory type class definition.

Its aim is to be used when you want to create directories that ARE :

 - NOT member or group directories (it means we don't want to store the entries
   within the acl_users 'cause it has nothing to do with those notions.
 - NOT aimed to be store in a LDAP Directory.

You might find that usefull when you'd like to store the directory within the
ZODB.
"""

from AccessControl import ClassSecurityInfo, Unauthorized
from Globals import InitializeClass, DTMLFile, Persistent
from OFS.SimpleItem import SimpleItem
from BTrees.OOBTree import OOBTree

import Acquisition

from Products.CMFCore.utils import getToolByName

from BaseDirectory import DirectorySchemaProperty
from MembersDirectory import GroupsDirectory

from zLOG import LOG, DEBUG

##################################################

class ZODBEntry(SimpleItem):
    """
    Class used to store an entry in the OOBTree.
    Persitance purpose + speed within the OOBTree
    """

    def wrapEntry(self, entry=None):
        """
        Just store the entry content.
        """
        if entry is not None:
            self.entry = entry

    def getEntry(self):
        """
        Encapsulation to keep that clean ;)
        """
        return self.entry

##################################################

ZODBIDPROPID    = "id_"
ZODBTITLEPROPID = "title_"

##################################################

class ZODBDirectory(GroupsDirectory):
    """
    ZODBDirectory
    Used to store directories within ZODB.
    """

    meta_type = "ZODB Directory"
    security  = ClassSecurityInfo()

    def __init__(self, id, **kw):
        """
        Constructor
        """

        GroupsDirectory.__init__(self, id, **kw)

        self.manage_properties(
            acl_write    = "python:manager",
            acl_create   = "manager",
            acl_delete   = "python:manager",
            entry_prop   = ZODBIDPROPID,
            display_prop = ZODBTITLEPROPID )

        self._schemakeys = (ZODBIDPROPID, ZODBTITLEPROPID)

        self._schema = \
        {ZODBIDPROPID: \
         DirectorySchemaProperty(ZODBIDPROPID,
                                 title="_dir_prop_Id_",
                                 alias=ZODBIDPROPID,
                                 acl_write="python:not property or nobody",
                                 process_write=\
                                 "python:property or malformed('Required')"),

         ZODBTITLEPROPID: \
         DirectorySchemaProperty(ZODBTITLEPROPID,
                                 title="_dir_prop_Name_",
                                 alias=ZODBTITLEPROPID),
        }

        # Used to store the dictionnaties containing the entries
        self._entries = OOBTree()

    def _createEntry(self, entry_id, entry):
        if not entry_id:
            return

        # Checking if the entry doesn't exist already
        if self._existEntry(entry_id):
            return

        # Creation of the ZODBEntry instance
        new_entry = ZODBEntry()
        new_entry.wrapEntry(entry)

        # Storing now.
        self._entries[entry_id] = new_entry

    def _existEntry(self, entry_id):
        """
        Check if the entry is already here
        """
        return self._entries.has_key(entry_id)

    def _writeEntry(self, entry_id, entry):
        """
        Updating the entry here.
        """
        if entry_id is not None:
            if self._entries.has_key(entry_id):
                entry[ZODBIDPROPID] = entry_id
                self._entries[entry_id].wrapEntry(entry)

    def _getEntry(self, entry_id):
        """
        Return the entry
        matching entry_id if found
        """
        if entry_id is not None:
            if self._entries.has_key(entry_id):
                return self._entries[entry_id].getEntry()
        return None

    def _deleteEntry(self, entry_id):
        """
        Deleting the entry from the OOBTree
        matching the entry_id if found
        """
        if entry_id is not None:
            if self._entries.has_key(entry_id):
                del self._entries[entry_id]

    def _searchEntry(self, search_options, **kw):
        """
        Seaching....
        """
        res_entries = []

        for key in self._entries.keys():
            nice = 1
            for schema_key in self._schemakeys:
                try:
                    form_key = kw.get(schema_key, '').lower()
                except AttributeError:
                    # tokens are stored here
                    form_key = kw.get(schema_key, '')[0].lower()

                field_value = self._entries[key].getEntry().get(schema_key, None)

                try:
                    find_string = field_value.lower().find(form_key)
                except AttributeError:
                    # tokens are stored here
                    try:
                        find_string = field_value[0].lower().find(form_key)
                    except:
                        # XXX expect what !! What is the use case ?
                        # find returns the first index of the substring if finds it , -1 if not.
                        find_string = -1

                if form_key != "" and \
                       field_value is not None and find_string == -1:
                    nice = 0
            if nice:
                res_entries.append(self._entries[key].getEntry())

        return res_entries

InitializeClass(ZODBDirectory)

########################################################

manage_addZODBDirectoryForm = DTMLFile('zmi/manage_addZODBDirectoryForm',
                                       globals())

########################################################

def manage_addZODBDirectory(dispatcher, id, title='', REQUEST=None):
    """
        Add a ZODB Directory
    """
    ob = ZODBDirectory(id, title=title)
    container = dispatcher.Destination()
    container._setObject(id, ob)
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)
